--------------------------------------------------
-- BonusScanner v1.1
-- by Crowley <crowley@headshot.de>
-- performance improvements by Archarodim
--
-- get the latest version here:
-- http://ui.worldofwar.net/ui.php?id=1461
-- http://www.curse-gaming.com/mod.php?addid=2384
--
-- Modified for HealPoints by Eridan
--------------------------------------------------

HEALPOINTSBS_PATTERN_SETNAME = "^(.*) %(%d/%d%)$";
HEALPOINTSBS_PATTERN_GENERIC_PREFIX = "^%+?(%d+)%%?(.*)$";
HEALPOINTSBS_PATTERN_GENERIC_SUFFIX = "^(.*)%+(%d+)%%?$";

HealPointsBS = {
	bonuses = {};
	bonuses_details = {};

  IsUpdating		    = false; -- not sure if this check is needed but who knows with multithreading...
  MinCheckInterval	    = 1;	 -- Minimum time to wait between each scan
  CheckIntervalCounter    = 0;	 -- counter, do not change
  CheckForBonusPlease	    = 0;	 -- The flag that when set makes BonusScanner scan the equipment and call the update function
  ShowDebug		    = false; -- tells when the equipment is scanned

	active = nil;
	temp = { 
		sets = {},
		set = "",
		slot = "",
		bonuses = {},
		details = {}
	};

	types = {
    "STR",      -- strength
		"INT", 			-- intellect
		"SPI", 			-- spirit
		"HEAL",			-- healing 
		"MANAREG",		-- mana regeneration per 5 sec.
		"MANA",			-- mana points

		-- Added
		"SPELLCRITRATING", -- spell critical strike rating
		"CASTINGREG", 		-- % mana regeneration while casting

		"CASTINGHOLYLIGHT", -- reduced casting time Holy Light

		"CASTINGFLASHHEAL", -- reduced casting time Flash Heal
		"GHEALRENEW",		-- Greater Heal also gives Renew effect
		"CHEAPERRENEW",		-- Cheaper Renew by %
		"DURATIONRENEW",    -- increased duration Renew

		"CASTINGREGROWTH", 	-- reduced casting time Regrowth
		"DURATIONREJUV",  	-- increased duration Rejuvenation
		"CHEAPERDRUID",		-- Cheaper druids spells by %
		"REFUNDHTCRIT",    	-- Refund of % mana on HT crits
    "DURATIONREGROWTH", -- increased duration Regrowth
    "IMPLIFEBLOOMBURST", -- increased effect from Lifebloom (Burst only)

		"REFUNDHEALINGWAVE",-- chance for mana refund on (Lesser) Healing Wave cast
		"JUMPHEALINGWAVE",	-- Healing Wave jumps to nearby targets
		"IMPCHAINHEAL",     -- increased effect from chain heal jumps
    "CHEAPERLESSERHEALINGWAVE", -- cheaper Lesser Healing Wave by %
    "CASTINGHEALINGWAVE", -- reduced casting time Healing Wave
		"CASTINGCHAINHEAL", -- reduced casting time Chain Heal

		"IMPFLASHOFLIGHT",  -- increased effect from Flash of Light
		"IMPBOL",  -- increased effect from Blessing of Light
		"IMPHOLYLIHG", -- increased effect from Holy Light

    "IMPHEALINGTOUCH", -- increased effect from Healing Touch
    "REFUNDHT", -- Refund x mana on HT casts
		"IMPREJUVENATION",  -- increased effect from Rejuvenation
    "IMPLIFEBLOOM", -- increased effect from Lifebloom (HoT only)

    "REFUNDLESSERHEALINGWAVE", -- Refund x mana on Lesser Healing Wave casts
    "IMPCHAINHEAL2", -- increased base effect of chain heal
		"IMPLESSERHEALINGWAVE", -- increased effect from Lesser Healing Wave
    "CHEAPERHEALINGWAVE", -- cheaper Healing Wave by %
    "IMPHEALINGWAVE", -- increased effect from Healing Wave
  };

	slots = {
		"Head",
		"Neck",
		"Shoulder",
		"Shirt",
		"Chest",
		"Waist",
		"Legs",
		"Feet",
		"Wrist",
		"Hands",
		"Finger0",
		"Finger1",
		"Trinket0",
		"Trinket1",
		"Back",
		"MainHand",
		"SecondaryHand",
		"Ranged",
		"Tabard",
	};
}

local slotConvert = {	
  INVTYPE_HEAD = "Head", 			
  INVTYPE_NECK = "Neck", 					
  INVTYPE_SHOULDER = "Shoulder", 				
  INVTYPE_BODY = "Shirt",
  INVTYPE_CHEST = "Chest", 		
  INVTYPE_ROBE = "Chest", 				
  INVTYPE_WAIST = "Waist", 					
  INVTYPE_LEGS = "Legs",
  INVTYPE_FEET = "Feet", 			
  INVTYPE_WRIST = "Wrist", 				
  INVTYPE_HAND = "Hands", 					
  INVTYPE_FINGER = "Finger0",
  INVTYPE_TRINKET = "Trinket0", 	
  INVTYPE_CLOAK = "Back", 				
  INVTYPE_WEAPON = "MainHand", 				
  INVTYPE_SHIELD = "SecondaryHand",
  INVTYPE_2HWEAPON = "MainHand", 	
  INVTYPE_WEAPONMAINHAND = "MainHand",	
  INVTYPE_WEAPONOFFHAND = "SecondaryHand",	
  INVTYPE_HOLDABLE = "SecondaryHand",	
  INVTYPE_RANGED = "Ranged",		
  INVTYPE_THROWN = "Ranged", 				
  INVTYPE_RANGEDRIGHT = "Ranged", 			
  INVTYPE_TABARD = "Tabard" ,
  INVTYPE_RELIC = "Ranged"
}; 
              
local function Debug(message) 
	-- A little debug function
  if (HealPointsBS.ShowDebug) then
    DEFAULT_CHAT_FRAME:AddMessage("HealPoints-Bonnus-Scanner: "..message, 0.5, 0.8, 1);
	end	
end

local function GetSlotBonuses(slotname) 
	local bonus, details;
	local bonuses = {};
	for bonus, details in pairs(HealPointsBS.bonuses_details) do
		if (details[slotname]) then
			bonuses[bonus] = details[slotname];
		end
	end
	return bonuses;
end

local function isEnchanted(frame) 
  for i=1,frame:NumLines() do
    local mytext = getglobal(frame:GetName().."TextLeft"..i);
    local text = mytext:GetText();
    local r, g, b, a = mytext:GetTextColor();
    if (r == 0 and g > 0.99 and b == 0 and
        string.sub(text, 0, string.len(ITEM_SPELL_TRIGGER_ONEQUIP)) ~= ITEM_SPELL_TRIGGER_ONEQUIP and
        string.sub(text, 0, string.len(ITEM_SET_BONUS)-2).."%s" ~= ITEM_SET_BONUS and 
        string.sub(text, 0, string.len(ITEM_SPELL_TRIGGER_ONUSE)) ~= ITEM_SPELL_TRIGGER_ONUSE and
        string.sub(text, 0, string.len(HEALPOINTSBS_SOCKET_MSG)) ~= HEALPOINTSBS_SOCKET_MSG and
        string.sub(text, 0, string.len(ITEM_SOCKET_BONUS)-2).."%s" ~= ITEM_SOCKET_BONUS) then
      return true;
    end
  end
  return false;
end

local function AddValue(effect, value) 
	local i,e;
	
	if (type(effect) == "string") then
		if (HealPointsBS.temp.bonuses[effect]) then
			HealPointsBS.temp.bonuses[effect] = HealPointsBS.temp.bonuses[effect] + value;
		else
			HealPointsBS.temp.bonuses[effect] = value;
		end
		
		if (HealPointsBS.temp.slot) then
			if (HealPointsBS.temp.details[effect]) then
				if (HealPointsBS.temp.details[effect][HealPointsBS.temp.slot]) then
					HealPointsBS.temp.details[effect][HealPointsBS.temp.slot] = HealPointsBS.temp.details[effect][HealPointsBS.temp.slot] + value;
				else
					HealPointsBS.temp.details[effect][HealPointsBS.temp.slot] = value;
				end
			else
				HealPointsBS.temp.details[effect] = {};
				HealPointsBS.temp.details[effect][HealPointsBS.temp.slot] = value;
			end
		end
	elseif (type(value) == "table") then
    for i,e in pairs(effect) do
      AddValue(e, value[i]);
    end
  else
    for i,e in pairs(effect) do -- TODO effect = nil
      AddValue(e, value);
    end
  end
end

local function CheckOther(line) 
	-- Last fallback for non generic enchants, like "Mana Regen x per 5 sec."
	local i, p, value1, value2;

	for i,p in pairs(HEALPOINTSBS_PATTERNS_OTHER) do
		value1, value2 = string.match(line, "^" .. p.pattern);
		if(value2) then -- Two effects
			if(p.value and p.value[2]) then
				AddValue(p.effect[1], p.value[1])
				AddValue(p.effect[2], p.value[2])
			else
				AddValue(p.effect[1], value1)
				AddValue(p.effect[2], value2)
			end
      return true;
		elseif(value1) then -- One effect
			if(p.value) then
				AddValue(p.effect, p.value)
			else
				AddValue(p.effect, value1)
			end
			return true;
		end
	end
	return false;
end


local function CheckToken(token, value) 
	-- Identifies simple tokens like "Intellect" and composite tokens like "Fire damage" and 
	-- add the value to the respective bonus. 
	-- returns true if some bonus is found
	if(HEALPOINTSBS_PATTERNS_GENERIC_LOOKUP[string.lower(token)]) then
		AddValue(HEALPOINTSBS_PATTERNS_GENERIC_LOOKUP[string.lower(token)], value);
		return true;
	end
	return false;
end

local function CheckGeneric(line) 
	-- Scans generic bonuses like "+3 Intellect" or "Arcane Resistance +4"
	local value, token, tmpStr;

  if (string.sub(line, 0, 2) == "|c") then -- fix for white enchants, stolen from ItemBonusLib
    line = string.sub(line, 11, -4)
  end
    
	-- split line at "/" for enchants with multiple effects
	local found = false;
	while(string.len(line) > 0) do
		local pos = string.find(line, "/", 1, true);
    local pos2 = string.find(line, "&", 1, true);
    local pos3 = string.find(line, ",", 1, true); -- TODO: Will this break something else?
		if (pos) then
			tmpStr = string.sub(line,1,pos-1);
			line = string.sub(line,pos+1);
		elseif (pos2) then
			tmpStr = string.sub(line,1,pos2-1);
			line = string.sub(line,pos2+1);    
    elseif (pos3) then
			tmpStr = string.sub(line,1,pos3-1);
			line = string.sub(line,pos3+1);        
    else
			tmpStr = line;
			line = "";
		end

		-- trim line

    tmpStr = string.gsub( tmpStr, "^%s+", "" ); -- Remove spaces from the beginning of the line
    tmpStr = string.gsub( tmpStr, "%s+$", "" ); -- Remove spaces from the end of the line
    tmpStr = string.gsub( tmpStr, "%.$", "" ); -- Remove . from the end of the line
    --DEFAULT_CHAT_FRAME:AddMessage("Line: *"..tmpStr.."* "..string.len(tmpStr),1,1,1);

		_, _, value, token = string.find(tmpStr, HEALPOINTSBS_PATTERN_GENERIC_PREFIX);
		if (not value) then
			_, _,  token, value = string.find(tmpStr, HEALPOINTSBS_PATTERN_GENERIC_SUFFIX);
		end
		if (token and value) then
			-- trim token
      token = string.gsub( token, "^%s+", "" );
      token = string.gsub( token, "%s+$", "" );
      token = string.gsub( token, "%.$", "" );
      --DEFAULT_CHAT_FRAME:AddMessage("Token: *"..token.."*",1,1,1);
	
			if(CheckToken(token,value)) then
				found = true;
			end
		end
	end
	return found;
end

local function CheckPassive(line) 
	-- Scans passive bonuses like "Set: " and "Equip: "
	local found = nil;
	local i, p, value;
	for i,p in pairs(HEALPOINTSBS_PATTERNS_PASSIVE) do
		_, _, value = string.find(line, "^" .. p.pattern);
		if (value) then
			AddValue(p.effect, value)
			found = 1;
			break; -- prevent duplicated patterns to cause bonuses to be counted several times
		end
	end
	if(not found) then
		Debug("\"" .. line .. "\"");
		CheckGeneric(line);
	end
end

local function ScanLine(line, includeSet, includeEnchant, isGreen) 
  --line = string.gsub(line, "^%s*(.-)%s*$", "%1");

	-- Check for "Equip: "
	if(string.sub(line,0,string.len(ITEM_SPELL_TRIGGER_ONEQUIP)) == ITEM_SPELL_TRIGGER_ONEQUIP) then
    CheckPassive(string.sub(line,string.len(ITEM_SPELL_TRIGGER_ONEQUIP)+2));

	-- Check for "Set: "
	elseif(includeSet ~= false and string.sub(line,0,string.len(ITEM_SET_BONUS)-2).."%s" == ITEM_SET_BONUS
			and HealPointsBS.temp.set ~= "" and not HealPointsBS.temp.sets[HealPointsBS.temp.set]) then
		HealPointsBS.temp.slot = "Set";
    local tmpStr = string.sub(line,string.len(ITEM_SET_BONUS)-1);
    --DEFAULT_CHAT_FRAME:AddMessage("Set bonus: -"..tmpStr.."-",1,1,1);    
		CheckPassive(tmpStr);

  elseif(string.sub(line,0,string.len(ITEM_SOCKET_BONUS)-2).."%s" == ITEM_SOCKET_BONUS and isGreen) then
    local tmpStr = string.sub(line,string.len(ITEM_SOCKET_BONUS)-1);
    --DEFAULT_CHAT_FRAME:AddMessage("Socket bonus: -"..tmpStr.."-",1,1,1);    
    
    local found = CheckGeneric(tmpStr);
    if (not found) then
      CheckOther(tmpStr);
    end;
    
	-- any other line (standard stats, enchantment, set name, etc.)
	else
		-- Check for set name
    local tmpStr
		_, _, tmpStr = string.find(line, HEALPOINTSBS_PATTERN_SETNAME);
		if(tmpStr) then
			HealPointsBS.temp.set = tmpStr;
		elseif (includeEnchant == true or isGreen == false) then
			local found = CheckGeneric(line);
			if(not found) then
				CheckOther(line);
			end;
		end
  end
end;


local function ScanTooltip(frame, includeSet, includeEnchant) 
	local lines = frame:NumLines();

	for i=2, lines, 1 do
		local tmpText = getglobal(frame:GetName().."TextLeft"..i);
		if (tmpText:GetText()) then
			local line = tmpText:GetText();      
      local r, g, b;
      r, g, b = tmpText:GetTextColor();
      local isGreen = (r == 0 and g > 0.99 and b == 0);
      ScanLine(line, includeSet, includeEnchant, isGreen);
		end
	end
end

local function ScanEquipment() 
	HealPointsBSTooltip:SetOwner(UIParent, "ANCHOR_NONE");

  Debug("Scanning Equipment has requested");

	HealPointsBS.temp.bonuses = {};
	HealPointsBS.temp.details = {};
	HealPointsBS.temp.sets = {};
	HealPointsBS.temp.set = "";

	local slotname, i;
	for i, slotname in pairs(HealPointsBS.slots) do
		local slotid, _ = GetInventorySlotInfo(slotname.. "Slot");
		local hasItem = HealPointsBSTooltip:SetInventoryItem("player", slotid);
	
		if (hasItem) then
			HealPointsBS.temp.slot = slotname;
			ScanTooltip(HealPointsBSTooltip, true, true);
			-- if set item, mark set as already scanned
			if (HealPointsBS.temp.set ~= "") then
				HealPointsBS.temp.sets[HealPointsBS.temp.set] = 1;
			end;
		end
	end
	HealPointsBS.bonuses = HealPointsBS.temp.bonuses;
	HealPointsBS.bonuses_details = HealPointsBS.temp.details;
end

local function ScanItem(frame, includeSet, includeEnchant) 
	HealPointsBS.temp.bonuses = {};
	HealPointsBS.temp.sets = {};
	HealPointsBS.temp.set = "";
	HealPointsBS.temp.slot = "";
	ScanTooltip(frame, includeSet, includeEnchant);
	return HealPointsBS.temp.bonuses;
end

-- Global functions
function HealPointsBS:update()
	-- Update function to hook into. 
	-- Gets called, when Equipment changes (after UNIT_INVENTORY_CHANGED)
end

function HealPointsBS:GetBonus(bonus)
	if(HealPointsBS.bonuses[bonus]) then
		if (type(HealPointsBS.bonuses[bonus]) == "string") then
			HealPointsBS.bonuses[bonus] = tonumber(HealPointsBS.bonuses[bonus]);
		end
    return HealPointsBS.bonuses[bonus];
	end
	return 0;
end

function HealPointsBS:getItemSlotBonuses(frame, itemLink)
  if (type(itemLink) == "string" ) then
    _,_, itemLink = string.find(itemLink, "item:(%d+):");
  end

  local _, _, _, _, _, type, subtype, _, itemEquipLoc = GetItemInfo(itemLink);	
  if (type == HEALPOINTSBS_GEM_TYPE and subtype == HEALPOINTSBS_GEM_SUBTYPE) then
    local itemBonuses = ScanItem(frame, false, true);
    return itemBonuses, { };
  end
  local slotName = slotConvert[itemEquipLoc];	
  if (slotName == nil) then
    return nil;
  end

  local isEnchanted = isEnchanted(frame);
  
  local itemBonuses = ScanItem(frame, false, true);
  local slotBonuses;   
  if (not isEnchanted) then
		local slotid, _ = GetInventorySlotInfo(slotName.. "Slot");
		local hasItem = HealPointsBSTooltip:SetInventoryItem("player", slotid);
    slotBonuses = ScanItem(HealPointsBSTooltip, false, false);
  else
    slotBonuses = GetSlotBonuses(slotName);
  end

  if (itemEquipLoc == "INVTYPE_2HWEAPON") then -- 2hander replaces 1hander + shield/offhand
    local slotBonuses2 = GetSlotBonuses("SecondaryHand");
    for index, value in pairs(slotBonuses2) do
      if (slotBonuses[index] == nil) then
        slotBonuses[index] = value;
      else
        slotBonuses[index] = slotBonuses[index] + value;
      end
    end
  end
  if (slotName == "SecondaryHand") then -- shield/offhand replaces 2hander
    local mainHandLink = GetInventoryItemLink("player",GetInventorySlotInfo("MainHandSlot"))
    if (mainHandLink ~= nil) then
      local _, _, itemCode = strfind(mainHandLink, "(%d+):")
      local _, _, _, _, _, _, _, _, loc = GetItemInfo(itemCode);
      if (loc == "INVTYPE_2HWEAPON") then
        slotBonuses = GetSlotBonuses("MainHand");			
      end
    end
  end
  if (slotName == "Finger0") then
    return itemBonuses, slotBonuses, GetSlotBonuses("Finger1");
  elseif (slotName == "Trinket0") then
    return itemBonuses, slotBonuses, GetSlotBonuses("Trinket1");
  else
    return itemBonuses, slotBonuses;
  end
end
  
-- Event functions
function HealPointsBS:OnLoad()
	this:RegisterEvent("PLAYER_ENTERING_WORLD");
	this:RegisterEvent("PLAYER_LEAVING_WORLD");
end

function HealPointsBS:OnEvent()
  Debug(event);

  if ((event == "UNIT_INVENTORY_CHANGED") and HealPointsBS.active) then
    HealPointsBS.CheckForBonusPlease = 1;
    return;
  end
	if (event == "PLAYER_ENTERING_WORLD") then
		HealPointsBS.active = 1;
		HealPointsBS.CheckForBonusPlease = 1;
		this:RegisterEvent("UNIT_INVENTORY_CHANGED");
    return;
	end
	if (event == "PLAYER_LEAVING_WORLD") then
		this:UnregisterEvent("UNIT_INVENTORY_CHANGED");
		return;
  end	
end

function HealPointsBS:OnUpdate(elapsed)
	-- The use of the <OnUpdate></OnUpdate> *feature* avoid freezes and lags caused by the useless repeated call of BonusScanner:ScanEquipment()...

  -- BonusScanner:Debug(elapsed);
  if (HealPointsBS.IsUpdating) then
    return;
  end

  HealPointsBS.IsUpdating = true;

  -- if the equipment has changed then check if we are allowed to test for bonuses
  if (HealPointsBS.CheckForBonusPlease == 1) then
    HealPointsBS.CheckIntervalCounter = HealPointsBS.CheckIntervalCounter + elapsed;

    -- if we have wait long enough then proceed...
    if (HealPointsBS.CheckIntervalCounter > HealPointsBS.MinCheckInterval) then
        HealPointsBS.CheckForBonusPlease = 2; -- means we are currently checking
        ScanEquipment(); -- scan the equiped items
        HealPointsBS:update();	  -- call the update function (for the mods using this library)
        if (HealPointsBS.CheckForBonusPlease ~= 1) then -- if no other update has been requested
          HealPointsBS.CheckForBonusPlease = 0;
        end
        HealPointsBS.CheckIntervalCounter = 0;
    end
  end

  HealPointsBS.IsUpdating = false;
end

